"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TaskGraphCreator = exports.nxCloudDistributedTasksRunner = void 0;
const stripJsonComments = require("strip-json-comments");
const fs_1 = require("fs");
const distributed_execution_api_1 = require("./distributed-execution.api");
const file_storage_1 = require("../../file-storage/file-storage");
const e2e_encryption_1 = require("../../file-storage/e2e-encryption");
const waiter_1 = require("../../../utilities/waiter");
const environment_1 = require("../../../utilities/environment");
const print_run_group_error_1 = require("../../error/print-run-group-error");
const create_no_new_messages_timeout_1 = require("../../../utilities/create-no-new-messages-timeout");
const metric_logger_1 = require("../../../utilities/metric-logger");
const error_reporter_api_1 = require("../../api/error-reporter.api");
const serializer_overrides_1 = require("../../../utilities/serializer-overrides");
const { output, workspaceRoot, getDependencyConfigs, getOutputs, unparse, Cache, } = require('../../../utilities/nx-imports');
class NoopLifeCycle {
    scheduleTask(task) { }
    startTask(task) { }
    endTask(task, code) { }
}
function runDistributedExecution(api, options, context, fileStorage, cache, runGroup, tasks) {
    return __awaiter(this, void 0, void 0, function* () {
        const nxjson = JSON.parse(stripJsonComments((0, fs_1.readFileSync)(`${workspaceRoot}/nx.json`).toString()));
        const stages = new TaskOrderer(context.projectGraph, getDefaultDependencyConfigs(nxjson, options)).splitTasksIntoStages(tasks);
        const id = yield api.start((0, distributed_execution_api_1.createStartRequest)(runGroup, stages, options));
        return yield processTasks(api, fileStorage, cache, options, id, tasks, context);
    });
}
function processTasks(api, fileStorage, cache, options, executionId, tasks, context) {
    return __awaiter(this, void 0, void 0, function* () {
        const processedTasks = {};
        const failIfNumberOfCompletedTasksDoesNotChangeIn30Mins = (0, create_no_new_messages_timeout_1.createNoNewMessagesTimeout)();
        const waiter = new waiter_1.Waiter();
        while (true) {
            if (environment_1.VERBOSE_LOGGING) {
                output.note({
                    title: 'Waiting...',
                });
            }
            yield waiter.wait();
            const r = yield api.status(executionId);
            if (environment_1.VERBOSE_LOGGING) {
                output.note({
                    title: `Status update`,
                    bodyLines: [
                        `executionId: ${executionId}`,
                        `executionStatus: ${r.executionStatus}`,
                        `number of completed tasks: ${r.completedTasks.length}`,
                        `error: ${r.criticalErrorMessage}`,
                    ],
                });
            }
            if (r.criticalErrorMessage) {
                output.error({
                    title: 'Distributed Execution Terminated',
                    bodyLines: ['Error:', r.criticalErrorMessage],
                });
                process.exit(1);
            }
            failIfNumberOfCompletedTasksDoesNotChangeIn30Mins(r.completedTasks.length);
            for (const t of r.completedTasks) {
                if (processedTasks[t.taskId])
                    continue;
                yield processTask(fileStorage, cache, context, options, tasks, t);
                waiter.reset();
                processedTasks[t.taskId] = true;
            }
            if (r.executionStatus === 'COMPLETED') {
                return { commandStatus: r.commandStatus, runUrl: r.runUrl };
            }
        }
    });
}
function processTask(fileStorage, cache, context, options, tasks, completedTask) {
    return __awaiter(this, void 0, void 0, function* () {
        if (environment_1.VERBOSE_LOGGING) {
            output.note({
                title: `Processing task ${completedTask.taskId}`,
            });
        }
        const matchingTask = tasks.find((tt) => completedTask.taskId === tt.id);
        if (!matchingTask) {
            throw new Error(`Found unknown task: ${completedTask.taskId}`);
        }
        if (environment_1.VERBOSE_LOGGING) {
            output.note({
                title: `Retrieving artifacts from ${completedTask.url}`,
            });
        }
        yield fileStorage.retrieve(completedTask.hash, completedTask.url, options.cacheDirectory || './node_modules/.cache/nx');
        const cachedResult = yield cache.get(Object.assign(Object.assign({}, matchingTask), { hash: completedTask.hash }));
        const outputs = getOutputs(context.projectGraph.nodes, matchingTask);
        if (environment_1.VERBOSE_LOGGING) {
            output.note({
                title: `Extracting artifacts`,
                bodyLines: outputs,
            });
        }
        yield cache.copyFilesFromCache(completedTask.hash, cachedResult, outputs);
        output.logCommand(getCommand(matchingTask));
        process.stdout.write(cachedResult.terminalOutput);
        output.addVerticalSeparator();
    });
}
function getCommand(task) {
    const config = task.target.configuration
        ? `:${task.target.configuration}`
        : '';
    return [
        'nx',
        'run',
        `${task.target.project}:${task.target.target}${config}`,
        (0, serializer_overrides_1.serializeOverrides)(task),
    ].join(' ');
}
const nxCloudDistributedTasksRunner = (tasks, options, context) => __awaiter(void 0, void 0, void 0, function* () {
    if (environment_1.VERBOSE_LOGGING) {
        output.note({
            title: 'Starting distributed command execution',
        });
    }
    options.lifeCycle = new NoopLifeCycle();
    const runGroup = (0, environment_1.getRunGroup)();
    if (!runGroup) {
        (0, print_run_group_error_1.printRunGroupError)();
        return process.exit(1);
    }
    const encryption = new e2e_encryption_1.E2EEncryption(environment_1.ENCRYPTION_KEY || options.encryptionKey);
    const errorReporter = new error_reporter_api_1.ErrorReporterApi(options);
    const fileStorage = new file_storage_1.FileStorage(encryption, errorReporter, true);
    const cache = new Cache(options);
    const api = new distributed_execution_api_1.DistributedExecutionApi(options);
    try {
        const r = yield runDistributedExecution(api, options, context, fileStorage, cache, runGroup, tasks);
        if (r.commandStatus === 0) {
            output.success({
                title: 'Successfully completed running the command.',
                bodyLines: [`See run details at ${r.runUrl}`],
            });
        }
        else {
            output.error({
                title: 'Command execution failed.',
                bodyLines: [`See run details at ${r.runUrl}`],
            });
        }
        yield (0, metric_logger_1.submitRunMetrics)(options);
        process.exit(r.commandStatus);
    }
    catch (e) {
        output.error({
            title: 'Unable to complete a run.',
            bodyLines: [e.message],
        });
        if (e.axiosException) {
            console.log(e.axiosException);
        }
        else {
            console.log(e);
        }
        try {
            yield api.completeRunGroupWithError(runGroup, `Main job terminated with an error: "${e.message}"`);
        }
        finally {
            process.exit(1);
        }
    }
});
exports.nxCloudDistributedTasksRunner = nxCloudDistributedTasksRunner;
class TaskOrderer {
    constructor(projectGraph, defaultTargetDependencies) {
        this.projectGraph = projectGraph;
        this.defaultTargetDependencies = defaultTargetDependencies;
    }
    splitTasksIntoStages(tasks) {
        if (tasks.length === 0)
            return [];
        const stages = [];
        const taskGraph = this.createTaskGraph(tasks);
        const notStagedTaskIds = new Set(tasks.map((t) => t.id));
        let stageIndex = 0;
        // Loop through tasks and try to stage them. As tasks are staged, they are removed from the loop
        while (notStagedTaskIds.size > 0) {
            const currentStage = (stages[stageIndex] = []);
            for (const taskId of notStagedTaskIds) {
                let ready = true;
                for (const dependency of taskGraph.dependencies[taskId]) {
                    if (notStagedTaskIds.has(dependency)) {
                        // dependency has not been staged yet, this task is not ready to be staged.
                        ready = false;
                        break;
                    }
                }
                // Some dependency still has not been staged, skip it for now, it will be processed again
                if (!ready) {
                    continue;
                }
                // All the dependencies have been staged, let's stage it.
                const task = taskGraph.tasks[taskId];
                currentStage.push(task);
            }
            // Remove the entire new stage of tasks from the list
            for (const task of currentStage) {
                notStagedTaskIds.delete(task.id);
            }
            stageIndex++;
        }
        return stages;
    }
    createTaskGraph(tasks) {
        const t = new TaskGraphCreator(this.projectGraph, this.defaultTargetDependencies);
        return t.createTaskGraph(tasks);
    }
}
class TaskGraphCreator {
    constructor(projectGraph, defaultTargetDependencies) {
        this.projectGraph = projectGraph;
        this.defaultTargetDependencies = defaultTargetDependencies;
    }
    createTaskGraph(tasks) {
        const graph = {
            roots: [],
            tasks: {},
            dependencies: {},
        };
        for (const task of tasks) {
            this.addTaskToGraph(task, graph);
            const dependencyConfigs = getDependencyConfigs(task.target, this.defaultTargetDependencies, this.projectGraph);
            if (!dependencyConfigs) {
                continue;
            }
            this.addTaskDependencies(task, dependencyConfigs, tasks, graph);
        }
        graph.roots = Object.keys(graph.dependencies).filter((k) => graph.dependencies[k].length === 0);
        return graph;
    }
    addTaskDependencies(task, dependencyConfigs, tasks, graph) {
        for (const dependencyConfig of dependencyConfigs) {
            if (dependencyConfig.projects === 'self') {
                for (const t of tasks) {
                    if (t.target.project === task.target.project &&
                        t.target.target === dependencyConfig.target) {
                        graph.dependencies[task.id].push(t.id);
                    }
                }
            }
            else if (dependencyConfig.projects === 'dependencies') {
                const seen = new Set();
                this.addDependencies(task.target.project, dependencyConfig.target, tasks, graph, task.id, seen);
            }
        }
    }
    addDependencies(project, target, tasks, graph, taskId, seen) {
        seen.add(project);
        const dependencies = this.projectGraph.dependencies[project];
        if (dependencies) {
            const projectDependencies = dependencies.map((dependency) => dependency.target);
            for (const projectDependency of projectDependencies) {
                if (seen.has(projectDependency)) {
                    continue;
                }
                const dependency = this.findTask({ project: projectDependency, target }, tasks);
                if (dependency) {
                    if (graph.dependencies[taskId].indexOf(dependency.id) === -1) {
                        graph.dependencies[taskId].push(dependency.id);
                    }
                }
                else {
                    this.addDependencies(projectDependency, target, tasks, graph, taskId, seen);
                }
            }
        }
    }
    findTask({ project, target }, tasks) {
        return tasks.find((t) => t.target.project === project && t.target.target === target);
    }
    addTaskToGraph(task, graph) {
        graph.tasks[task.id] = task;
        graph.dependencies[task.id] = [];
    }
}
exports.TaskGraphCreator = TaskGraphCreator;
function getDefaultDependencyConfigs(nxJson, runnerOptions) {
    var _a, _b;
    const defaults = (_a = nxJson.targetDependencies) !== null && _a !== void 0 ? _a : {};
    const strictlyOrderedTargets = runnerOptions
        ? (_b = runnerOptions.strictlyOrderedTargets) !== null && _b !== void 0 ? _b : ['build']
        : [];
    // Strictly Ordered Targets depend on their dependencies
    for (const target of strictlyOrderedTargets) {
        defaults[target] = defaults[target] || [];
        defaults[target].push({
            target,
            projects: 'dependencies',
        });
    }
    return defaults;
}
//# sourceMappingURL=distributed-execution.runner.js.map